/*****************************************************************************
Deng, Minjie and Fang, Min (2021) 
``Debt Maturity and Investment Responses to Monetary Policy''.

Program: 00_Do-Construct_distance-to-default
Purpose: this file estimates firm-level "distance to default" based on Gilchrist
and Zakrajsek (2012) and Bharath and Shumway (2008).

Code Modified from Ottonello and Winberry (2021) ECTA Paper
``Financial Heterogeneity and the Investment Channel of Monetary Policy'' 
Program: construct_d2default.do
*****************************************************************************/

clear all
set more off
set scheme s1color
capture log close


****************************************
* 1. Build dataset
****************************************
/*
Data sources and inputs:
CRSP: market cap (firm equity) and value-weighted stock return (w/ dividends)
Compustat: short-term and long-term debt
CRSP-Compustat Merge (CCM): gvkey-to-permco concordance
FRED: 1yr constant maturity nominal treasury yield

Output: dd_data.dta
*/


* Import CRSP for daily market cap (firm equity) and stock returns
// Date range: 1980-01-01 to 2014-12-31 (includes a few extra years to accommodate lags)
// Company codes: permco (search entire database)
// Select variables: price, number of shares outstanding
use "2_data_raw/crsp_raw.dta", clear
drop PERMNO
rename date dated
rename PERMCO permco
rename PRC prc
rename SHROUT shrout
/* Negative price denotes daily average instead of end-of-day price. Flip to include those observations. */
gen mktcap = abs(prc)*(shrout*1e3)/1e6 // market cap in millions
/* Collapse to firm-date (PERMCO-DATED). Market cap is sum of market cap 
	(mktcap) of all securities (PERMNO) under single firm-date observation. */
collapse (sum) mktcap, by(permco dated)

compress
save "5_temp/crsp.dta", replace


* Import CCM sample
// Date range: 1983-01 to 2014-12 (sample period in final 
// sample in regressions_firmlevel_data_final.dta)
// Company codes: gvkey (search entire database)
// Linking options: LC LU (Fiscal period end date must be within link date range)
// Query variables: Historical CRSP PERMCO Link to COMPUSTAT Record
use "2_data_raw/ccm_raw.dta", clear
rename GVKEY gvkey
rename LPERMCO lpermco
keep gvkey lpermco datacqtr
duplicates drop
rename lpermco permco
// Clean dates
gen year = substr(datacqtr,1,4)
gen quarter = substr(datacqtr,-1,1)
destring year quarter, replace
gen dateq = yq(year,quarter)
format dateq %tq
drop if mi(dateq)
drop datacqtr year quarter
destring gvkey, replace
// Drop duplicate observations (currently no obs from GVKEY-DATEQ and 213 from PERMCO-DATEQ)
foreach var in gvkey permco {
	duplicates tag `var' dateq, gen(dup_id)
	drop if dup_id
	drop dup_id
}

compress
save "5_temp/ccm.dta", replace


* Import FRED data
use "5_temp/FRED_tyields.dta", clear
keep DGS1 dated
// Linearly interpolate missing dates
gen miDGS1 = mi(DGS1)
tsset dated
tsfill
replace miDGS1 = 0 if mi(miDGS1)
ipolate DGS1 dated, gen(DGS1_ipolate)
replace DGS1 = DGS1_ipolate if miDGS1 // 648 observations interpolated
drop if mi(DGS1)
drop miDGS1 DGS1_ipolate
// Housekeeping
rename DGS1 trate1yr
compress
save "5_temp/trate1yr.dta", replace


* Construct debt series from original Compustat file [longer sample with only key variables to save space]
use "2_data_raw/compustat_quarterly_raw_long.dta", clear
keep gvkey datacqtr dlcq dlttq
// Clean dates
gen year = substr(datacqtr,1,4)
gen quarter = substr(datacqtr,-1,1)
destring year quarter, replace
gen dateq = yq(year,quarter)
format dateq %tq
drop if mi(dateq)
drop datacqtr year quarter
destring gvkey, replace
// Generate debt
gen debt = dlcq + 0.5*dlttq // GZ assumes debt = short-term + 1/2 long-term
keep gvkey dateq debt
drop if mi(debt)

compress
save "5_temp/debt.dta", replace


* Merge final data
use "5_temp/crsp.dta", clear
gen dateq = qofd(dated)
format dateq %tq
mer m:1 permco dateq using "5_temp/ccm.dta", nogen keep(3)
// Merge on debt from Compustat
mer m:1 gvkey dateq using "5_temp/debt.dta", nogen keep(3)
// Merge on risk-free rate
mer m:1 dated using "5_temp/trate1yr.dta", nogen keep(3)


* Housekeeping
// Clean variable names
drop permco
rename (mktcap debt trate1yr) (E D r)
order gvkey dated dateq
sort gvkey dated
// Finalize units
// vwretd already in decimal form
// Equity and Debt already in millions
replace r = r/100
// Nominal interest rate already in percent
// Label
lab var gvkey "Firm ID"
lab var dated "Date (daily)"
lab var dateq "Date (quarterly)"
lab var E "Equity (mkt cap)"
lab var D "Debt"
lab var r "1yr treasury constant maturity"

compress
save "5_temp/d2default_data.dta", replace


****************************************
* 2. Construct dd
****************************************

// Set length of rolling window for calculating avg and variance of returns
local T = 250


* Load final data constructed in build section
use "5_temp/d2default_data.dta", clear
// Sample cleaning to ensure distance to default can be calculated
drop if E==0 // Can't solve for V with E==0
drop if D==0 // Can't solve for V with D==0
// Identify end-of-quarter (eoq) observations
sort gvkey dateq dated
by gvkey dateq: gen eoq = _n==_N
// Ignore eoq observations with fewer than T prior observations
by gvkey: gen count = _n
replace eoq = 0 if count<`T'
drop count


* Split sample into 100,000 observation increments
local files = ceil(_N/100000)
forvalues s=1/`files' {
	display "File `s' of `files'"
	local start = (`s'-1)*100000+1
	local end = `s'*100000+`T'
	if `s'==`files' local end = _N
	preserve
		keep in `start'/`end'
		save "5_temp/d2d_temp/d2default_`s'.dta", replace
	restore
}


forvalues s=1/`files' {
use "5_temp/d2d_temp/d2default_`s'.dta", clear
// Replace eoq if in the first T observations because the previous file will have captured them
replace eoq = 0 in 1/`T'

// Algorithm setup
local tol = 1e-3
local iter = 100
gen count = _n // Observation counter
xtset gvkey count
// Keep track of run time
timer clear
timer on 1
// Save total iteration count and running tally to display during execution
egen eoq_total = total(eoq)
local eoq_count = eoq_total
drop eoq_total
local tally = 0


* Algorithm
// 1. Guess an initial value for V (value of firm assets)
gen resolution = .
gen V = .
gen V0 = E + D
gen V_final = .
gen muV_final = .
gen sigV_final = .

local obs = _N
forvalues i=1/`obs' {
if eoq[`i']==1 { // Only run if observation is eoq
	local ++tally
	local it = 1
	local loopend = 0
	if mod(`tally',25)==0 {
	display "File `s', observation `tally' of `eoq_count'."
	}
	qui {
	replace V = V0 if inrange(count,`i'-`T',`i')
	
	while `loopend'==0 {
		// 2. Compute the mean and variance of the rate of return using V over a rolling window T
		gen dlnV = log(V) - log(L.V)
		su dlnV if inrange(count,`i'-`T',`i')
		gen sigV = sqrt(250)*`r(sd)' if inrange(count,`i'-`T',`i') // annualize with 250 trading days per year
		gen muV = 250*`r(mean)' if inrange(count,`i'-`T',`i')
		// 3. Compute V_new
		gen delta1 = (1/sigV^2)*(log(V/D)+(r+0.5*sigV^2))
		gen delta2 = delta1 - sigV
		gen N1 = normal(delta1)
		gen N2 = normal(delta2)
		gen V_new = (E+exp(-r)*D*N2)/N1
		// 4. Check if V_new - V is small
		gen dV = abs(V_new - V)
		egen max_dV = max(dV)
		gen V_new_mi = log(V_new) - log(L.V_new)
		su V_new_mi
		local V_new_obs = `r(N)'
		if max_dV < `tol' {
			display "V converged."
			su dlnV
			replace resolution = 1 if count==`i'
			replace sigV_final = sqrt(250)*`r(sd)' if count==`i'
			replace muV_final = 250*`r(mean)' if count==`i'
			replace V_final = V if count==`i'
			local loopend = 1
		}
		if max_dV>1e20 {
			display "V diverged."
			replace resolution = 2 if count==`i'
			replace sigV_final = . if count==`i'
			replace muV_final = . if count==`i'
			replace V_final = . if count==`i'
			local loopend = 1
		}
		if `V_new_obs'==0 {
			display "V could not be calculated."
			replace resolution = 3 if count==`i'
			replace sigV_final = . if count==`i'
			replace muV_final = . if count==`i'
			replace V_final = . if count==`i'
			local loopend = 1
		}
		if `it'==`iter' {
			display "Iterations exhausted."
			replace resolution = 4 if count==`i'
			replace sigV_final = . if count==`i'
			replace muV_final = . if count==`i'
			replace V_final = . if count==`i'
			local loopend = 1
		}
		replace V = V_new
		drop dlnV sigV muV delta1 delta2 N1 N2 V_new dV max_dV V_new_mi
		local ++it
	}
	}
}
}

keep if eoq

compress
save "5_temp/d2d_temp/d2default_out`s'.dta", replace
*rm "5_temp/d2d_temp/d2default_`s'.dta"
}


* Append files
clear
forvalues i=1/`files' {
	if mod(`i',20)==0 display `i'
	append using "5_temp/d2d_temp/d2default_out`i'.dta"
*	rm "5_temp/d2d_temp/d2default_out`i'.dta"
}
// Remove duplicates with lower count
drop if count<=`T'
duplicates report gvkey dated
assert `r(unique_value)'==`r(N)'
// Housekeeping
lab def lab_resolution 1 "Converged" 2 "Diverged" 3 "V could not be calculated" 4 "Iterations exhausted"
lab val resolution lab_resolution

drop eoq count V V0
rename *_final *

compress
save "5_temp/d2default_outall.dta", replace


* Construct distance-to-default
gen d2default = (ln(V/D) + (muV - 0.5*sigV^2))/sigV

// Housekeeping
keep gvkey dateq d2default

compress
save "3_data_constructed/d2default.dta", replace





