/*****************************************************************************
Deng, Minjie and Fang, Min (2021) 
``Debt Maturity and Investment Responses to Monetary Policy''.

Program: 02_Do-Clean_Compustat_Quarterly_Sample
Purpose: this file construct sample for regression
*****************************************************************************/

clear all
set more off
set scheme s1color
capture log close

*-----------------------------------------------------------------------------------
*** Clean Quarterly
clear
use "2_data_raw/compustat_quarterly_raw", clear
drop datadate
keep if loc == "USA"
gen year=substr(datacqtr,1,4)
gen quarter=substr(datacqtr,6,1)
destring year quarter, replace
drop indfmt consol popsrc datafmt curcdq datacqtr datafqtr costat loc 
sort gvkey fyearq fqtr 
destring gvkey, replace


*-- Calculate Sales Growth and Asset Growth
gen sale_g=saleq/saleq[_n-1]-1 if gvkey==gvkey[_n-1]
gen asset_g=atq/atq[_n-1]-1 if gvkey==gvkey[_n-1]

*-- Calculate Investment by Almeida et al. (2012)
gen delta_k=capxy-capxy[_n-1] if gvkey==gvkey[_n-1] & fqtr-fqtr[_n-1]==1
replace delta_k=capxy if fqtr==1
gen inv=delta_k/ppentq
drop delta_k capxy

*-- Calculate Leverage follow Ottonello-Winberry (2017)
gen leverage=(dlcq+dlttq)/atq
gen maturity=dlttq/(dlcq+dlttq)

*-- Calculate Profitbility
gen profitability=oibdpq/atq

*-- Calculate Earnings Volatility
gen ROA=abs(niq/atq)
gen vol_avg= ROA+ROA[_n-1]+ROA[_n-2]+ROA[_n-3]+ROA[_n-4] if gvkey==gvkey[_n-4]
replace vol_avg=vol_avg/5

*-- Calculate Equity issurance follow "Target Behavior and Financing: How Conclusive is the Evidence?"
gen equity=lseq-ltq
gen delta_equity=equity-equity[_n-1] if gvkey==gvkey[_n-1]  
gen delta_req=req-req[_n-1] if gvkey==gvkey[_n-1]
gen nei= delta_equity - delta_req
* the units are millions

*- Calculate debt issurance follow "Target Behavior and Financing: How Conclusive is the Evidence?"
gen delta_atq=atq-atq[_n-1] if gvkey==gvkey[_n-1] 
gen ndi=delta_atq-delta_equity

*- Liquidity
gen liquidity = cheq/atq

*-----------------------------------------------------------------------------------
*** Sample Selection
*-- Sample Selection follow Almeida et al.(2012)

*-- Fiscal year mismatch
keep if fyr==3 |fyr==6 |fyr==9 |fyr==12

*-- exclude financial firms (SIC codes 6000 to 6999)
*-- exclude utilities (SIC codes 4900 to 4999)
*-- exclude NGO, Government enterprises (SIC codes 8000s, 9000s)
destring sic, replace
drop if sic>=8000
drop if sic>=6000 & sic<=6999
drop if sic>=4900 & sic<=4999

*-- Growth Assets > 100% in a quarter
drop if asset_g>1

*-- Missing sales number
drop if saleq==.

*-- Negative sales
drop if saleq<0

*-- Growth of Sales > 100% in a quarter
drop if sale_g>1

*-- Cash is greater than Assets
drop if cheq>atq

*-- PPE is greater than Assets
drop if ppentq>atq

*-- ATQ <10
drop if atq<10

*-- Missing DD1 or Missing DLTT
drop if dlcq==. | dlttq==.

*-- DD1 + DLTT > Assets
drop if dlc+dlttq>atq

*-- DLTT over Assets < 5%
drop if dltt<0.05*at

save "5_temp/sample_temp.dta", replace 


*-----------------------------------------------------------------------------------
*Merge Shocks, Aggregates, Age, D2D
*-----------------------------------------------------------------------------------
gen dateq = yq(year,quarter)

*** Merge shocks
sort gvkey year quarter
merge m:1 year quarter using "3_data_constructed/shocks.dta"
keep if _merge==3
drop _merge

*** Merge aggregate variables
merge m:1 year quarter using "2_data_raw/aggregate_controls_raw.dta"
keep if _merge==3
drop _merge
sort gvkey year quarter

*** Merge distance to default series - low match because sample based on final sample
merge 1:1 gvkey dateq using "3_data_constructed/d2default.dta", nogen keep(1 3) keepus(d2default)
sort gvkey year quarter

*** Merge age
* Record first quarter in which each firm appears to construct age variable later
merge m:1 gvkey using "3_data_constructed/DS_yearinc.dta", nogen keep(1 3)
sort gvkey year quarter
gen age = dateq - inc_dateq


*-----------------------------------------------------------------------------------
*Final Adjustment
*-----------------------------------------------------------------------------------
*-- Sector 2-digit
gen sic2=int(sic/100)

*-- Time
gen time=yq(year,quarter)
format time %tq
xtset gvkey time
sort gvkey time
*-- Generate Total Debt
gen dlq=dlcq+dlttq
gen borrow=d.dlq/l.dlq
gen longborrow=d.dlttq/l.dlq
gen shortborrow=d.dlcq/l.dlq
gen changemat=d.maturity

*-- Extreme Values
centile(leverage), centile(0.5, 99.5)
drop if ( leverage < r(c_1) | leverage > r(c_2) )

centile(inv), centile(0.5, 99.5)
drop if ( inv < r(c_1) | inv > r(c_2) )

centile(maturity), centile(0.5, 99.5)
drop if ( maturity < r(c_1) | maturity > r(c_2) )

sum leverage inv maturity borrow longborrow shortborrow
codebook leverage inv maturity

*-------------------------------------------------------------------------------
*--- Adjust independent variables by STD, or very normal values i.e., 25bps for monetary policy shocks
sum tsum wsum tsmo wsmo lev maturity
codebook tsum lev maturity
replace tsum=tsum/25
replace wsum=wsum/25
replace tsmo=tsmo/25
replace wsmo=wsmo/25
replace mp_jk=mp_jk/25
*--- Sign Flip
replace tsum = -tsum
replace wsum = -wsum
replace tsmo = -tsmo
replace wsmo = -wsmo
replace mp_jk= -mp_jk
*de=mean, which is now coefficient for tightq is for an average leverage, average maturity firm
replace inv = inv*100
replace borrow = borrow*100
replace longborrow = longborrow*100
replace shortborrow = shortborrow*100


save "3_data_constructed/sample.dta", replace 








